# Copyright (C) 2020 Ludvig Larsson aka Valmond ludviglarsson@protonmail.com
# This code is licensed under the terms of the GNU Public License, version 3 or later (GPL-3.0-or-later)

import sys, os
import json
import database_query as dq
import rsa_helper as rh
from configuration import get_configuration_value
import sqlite3
from sqlite3 import Error


"""
Creates a link file from existing data (that was created with the insert_data.py)
"""

def print_help():
    print('Usage:')
    print('')
    print('python3 get_link_from_data.py dataname')
    print('python3 get_link_from_data.py --testrun')

    sys.exit()


# Returns the connection so we don't need to open it again, or None if it fails
def generate_database(database_file_name):
    # delete any old db
    if os.path.isfile(database_file_name):
        os.remove(database_file_name)

    # Generate the database
    try:
        connection = sqlite3.connect(database_file_name)
    except Error:
        print("Could not create the (database) link:")
        print(Error)
        return None

    connection.execute('''CREATE TABLE information
             (
              name                  TEXT,
              originalfilename      TEXT,
              file_version          INTEGER DEFAULT 1,
              key                   TEXT,
              iv                    INTEGER,
              data_rsa_pub          TEXT
               );''')

    # The addresses where we'll try to fetch the data
    # address_list.append({'rsa_pub': pub, 'ip': ip, 'port': port})
    connection.execute('''CREATE TABLE address_list
             (
              rsa_pub       TEXT,
              ip            TEXT,
              port          TEXT
             );''')

    # This is n ot used (yet, is it useful?)
    #connection.execute('''CREATE TABLE encrypted_address_list
    #         (
    #          name              CHAR,
    #          originalfilename       CHAR,
    #          key     TEXT,
    #          iv INTEGER,
    #          data_rsa_pub TEXT
    #           );''')

    return connection


def set_information_in_database(connection, name, original_file_name, aes_key, aes_iv, data_rsa_pub_key):
    cursor = connection.cursor()
    cursor.execute('INSERT INTO information (name, originalfilename, key, iv, data_rsa_pub) VALUES(?, ?, ?, ?, ?)', (name, original_file_name, aes_key, aes_iv, data_rsa_pub_key))
    connection.commit()


def add_address_to_database(connection, rsa_pub_key, ip, port):
    cursor = connection.cursor()
    cursor.execute('INSERT INTO address_list (rsa_pub, ip, port) VALUES(?, ?, ?)', (rsa_pub_key, ip, port))
    connection.commit()


# Takes 'Name' entry from global database and generates a link named 'name'.10f which contains
# This link is a SQLite database and contains two tables;
# One for global information: IP, PORT, NAME, ORIGFILENAME, AES&IV, RSA_PUB
# (AES is stored in 64 bytes hex string)
# One address entry for each finger storing the data we are generating the link for.
# If version == 0 then the latest version for this data will be used
# if exclude_our_address == True then we will not add our own address to the link, only other nodes links
def create_and_save_link_database(name, data_version=0, exclude_our_address=False):
    row = dq.get_link_data_by_name(name)
    key_pair = dq.get_rsa_hex_keys()

    if row is None or key_pair is None:
        if row is None:
            print('#Error: Link not found in datbase.')
        if key_pair is None:
            print('#Error: RSA keys not found in datbase.')
        return False

    # Get data id from global database:
    data_id = dq.get_link_dataid_by_name(name)

    # Get a list of finger id
    fingers = dq.get_all_fingers(data_id)
    if len(fingers) == 0:
        print('This data have not yet been uploaded elsewhere!')

    # get the latest data_version
    if data_version == 0:
        data_version = dq.get_highest_version(data_id)

    # Get fingers with correct versions:
    if data_version > 0:
        fingers = dq.get_fingers_by_version(data_id, data_version)
        if len(fingers) == 0 and exclude_our_address:
            print('The new version of the data (version: ' + str(data_version) + ') has not yet been uploaded elsewhere!')
            return False

    connection = generate_database(name + '.10f')
    if connection is None:
        return False

    prv, pub = key_pair

    # Update the database link with 'my' address information
    if not exclude_our_address:
        add_address_to_database(connection, pub, get_configuration_value('ip'), get_configuration_value('port'))

    print("Fingers:", len(fingers), flush=True)

    fingers_len = len(fingers)
    for n in range(0, fingers_len):
        finger_id = fingers[n]
        # add to the address list:
        node_id,_ = dq.get_finger_data(finger_id)
        if node_id != 0:
            # get rsa_pub, ip, port from node and add it to the address list
            node_data = dq.get_node_data(node_id)
            # Update the link with address information
            add_address_to_database(connection, node_data[2], node_data[0], node_data[1])

    # data_rsa is, together with the name, the unique identifier for this data (e.g the rsa_pub key from the node owning the data)
    _, data_rsa_pub = rh.get_rsa_keypair(True)
    data_rsa_pub = data_rsa_pub.exportKey('DER').hex()

    # Update the link with information
    set_information_in_database(connection, name, row[1], row[2], row[3], data_rsa_pub)

    print('Link created:', name + '.10f')

# Takes 'Name' entry from database and generates a link named 'name'.10f which contains
# IP, PORT, NAME, ORIGFILENAME, AES&IV, RSA_PUB
# AES is stored in 64 bytes hex string
# the link is saved on disc.
# DEPRECATED
def create_and_save_link(name):
    print('Creating link:', name)

    row = dq.get_link_data_by_name(name)
    keypair = dq.get_rsa_hex_keys()

    if row is not None and keypair is not None:
        prv, pub = keypair
        # Get our local ip:port address (for now it's manual text file)

        # todo: add the IP:PORT:RSA_PUB for all the fingers pointing to this data

        # Create the json link:
        address_list = []
        # Add 'my' address
        ip = get_configuration_value('ip', '127.0.0.1')
        port = get_configuration_value('port', 10000)
        address_list.append({'rsa_pub': pub, 'ip': ip, 'port': port})

        # todo: add fingers addresses:
        # Get data id:
        dataid = dq.get_link_dataid_by_name(name)

        print("FINGERS:", name, 'dataid:', dataid, flush=True)


        # Get a list of finger id
        fingers = dq.get_all_fingers(dataid)

        if fingers == 0:
            print('This link have not yet been uploaded elsewhere!')

        print("Fingers:", len(fingers), flush=True)

        fingers_len = len(fingers)
        print("FINGERS 0 len=", fingers_len, flush=True)
        for n in range(0, fingers_len):
            print("FINGERS 1", n, flush=True)
            fingerid = fingers[n]
            print("FINGERS 2", type(fingerid), flush=True)
            # add to the address list:
            print("FINGERS 3 fingerid=", fingerid, flush=True)
            nodeid, dataid_check = dq.get_finger_data(fingerid)
            print("FINGERS 4 dataid:", dataid, "dataid_check:", dataid_check, "nodeid:", nodeid, flush=True)

            if dataid is not dataid_check:
                print("#Error in SQL, dataid is not dataid_check")
            if nodeid != 0:
                # get rsa_pub, ip, port from node and add it to the address list
                node_data = dq.get_node_data(nodeid)
                address_list.append({'rsa_pub': node_data[2], 'ip': node_data[0], 'port': node_data[1]})

        # Get the new, encrypted, address:
        encrypted_address = dq.get_encrypted_addresses(dataid)


        print("FINGERS X", flush=True)
        # generate link data
        # data_rsa is, together with the name, the unique identifier for this data (e.g the rsa_pub key from the node owning the data)
        _, data_rsa_pub = rh.get_rsa_keypair(True)
        data_rsa_pub = data_rsa_pub.exportKey('DER').hex()

        d = {'name': name, 'originalfilename': row[1], 'key': row[2], 'iv': row[3], 'data_rsa_pub': data_rsa_pub,
             'address_list': address_list, 'encrypted_address_list': encrypted_address}

        # todo: sign the link

        s = json.dumps(d)
        beautify = True  # Insert linefeeds in json
        if beautify:
            s = s.replace("}", "}\n")
        with open(name+'.10f', 'w') as f:
            f.write(s)

        print('Link created.')
    else:
        if row is None:
            print('#Error: Link not found in datbase.')
        if keypair is None:
            print('#Error: RSA kaypair not found in datbase.')


# Test case:
def test():
    print('[get_link_from_data.py]')
    create_and_save_link('test')
    create_and_save_link('image')



# Call the main function with the command line arguments
if __name__ == '__main__':

    if len(sys.argv) == 1:
        create_and_save_link_database('worigatenfingerssite')
        print_help()

    if sys.argv[1] == '--testrun':
        test()
    else:
        create_and_save_link_database(sys.argv[1])
